package org.pabloid.ui;

import java.util.*;
import java.io.*;
import javax.microedition.lcdui.*;
import org.pabloid.io.*;
import org.pabloid.util.regexp.*;
import org.pabloid.util.*;

/**
 * The <code>FileBrowser</code> custom component lets the user list files and
 * directories. It's uses FileConnection Optional Package (JSR 75). The FileConnection
 * Optional Package APIs give J2ME devices access to file systems residing on mobile devices,
 * primarily access to removable storage media such as external memory cards.
 * @author breh
 */
public class FileBrowser extends PList implements CommandListener, Comparator
  {

    private static final byte imDir[] =
      {
        (byte) 0x89, (byte) 0x50, (byte) 0x4E, (byte) 0x47, (byte) 0x0D, (byte) 0x0A,
        (byte) 0x1A, (byte) 0x0A, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x0D,
        (byte) 0x49, (byte) 0x48, (byte) 0x44, (byte) 0x52, (byte) 0x00, (byte) 0x00,
        (byte) 0x00, (byte) 0x0A, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x0A,
        (byte) 0x08, (byte) 0x06, (byte) 0x00, (byte) 0x00, (byte) 0x01, (byte) 0xFA,
        (byte) 0x35, (byte) 0xFF, (byte) 0x2B, (byte) 0x00, (byte) 0x00, (byte) 0x00,
        (byte) 0x04, (byte) 0x67, (byte) 0x41, (byte) 0x4D, (byte) 0x41, (byte) 0x00,
        (byte) 0x00, (byte) 0xB1, (byte) 0x8F, (byte) 0x0B, (byte) 0xFC, (byte) 0x61,
        (byte) 0x05, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x06, (byte) 0x62,
        (byte) 0x4B, (byte) 0x47, (byte) 0x44, (byte) 0x00, (byte) 0xFF, (byte) 0x00,
        (byte) 0xFF, (byte) 0x00, (byte) 0xFF, (byte) 0xA0, (byte) 0xBD, (byte) 0xA7,
        (byte) 0x93, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x09, (byte) 0x70,
        (byte) 0x48, (byte) 0x59, (byte) 0x73, (byte) 0x00, (byte) 0x00, (byte) 0x0B,
        (byte) 0x13, (byte) 0x00, (byte) 0x00, (byte) 0x0B, (byte) 0x13, (byte) 0x01,
        (byte) 0x00, (byte) 0x9A, (byte) 0x9C, (byte) 0x18, (byte) 0x00, (byte) 0x00,
        (byte) 0x00, (byte) 0x7D, (byte) 0x49, (byte) 0x44, (byte) 0x41, (byte) 0x54,
        (byte) 0x18, (byte) 0xD3, (byte) 0x8D, (byte) 0x8F, (byte) 0x6D, (byte) 0x0D,
        (byte) 0xC3, (byte) 0x30, (byte) 0x0C, (byte) 0x05, (byte) 0xCF, (byte) 0xD1,
        (byte) 0x90, (byte) 0x14, (byte) 0x4B, (byte) 0x29, (byte) 0x84, (byte) 0x42,
        (byte) 0x39, (byte) 0xB5, (byte) 0x10, (byte) 0x12, (byte) 0x0A, (byte) 0xA3,
        (byte) 0x92, (byte) 0x52, (byte) 0xB9, (byte) 0xFD, (byte) 0x48, (byte) 0xA3,
        (byte) 0x4E, (byte) 0xEB, (byte) 0xBE, (byte) 0x4E, (byte) 0x7A, (byte) 0x72,
        (byte) 0x6C, (byte) 0x3F, (byte) 0x47, (byte) 0x76, (byte) 0xA8, (byte) 0x00,
        (byte) 0xA4, (byte) 0xBA, (byte) 0x85, (byte) 0xEC, (byte) 0x21, (byte) 0x65,
        (byte) 0x45, (byte) 0x6C, (byte) 0xF4, (byte) 0x97, (byte) 0xAD, (byte) 0xAB,
        (byte) 0x27, (byte) 0x4A, (byte) 0xD8, (byte) 0x10, (byte) 0x20, (byte) 0x31,
        (byte) 0x78, (byte) 0xF6, (byte) 0xD9, (byte) 0xB0, (byte) 0xAC, (byte) 0x78,
        (byte) 0x76, (byte) 0x0F, (byte) 0xF2, (byte) 0x0C, (byte) 0x51, (byte) 0xD6,
        (byte) 0x3E, (byte) 0x0B, (byte) 0x90, (byte) 0x17, (byte) 0x03, (byte) 0x38,
        (byte) 0x3F, (byte) 0x04, (byte) 0xA8, (byte) 0xF7, (byte) 0x1E, (byte) 0xD3,
        (byte) 0xEB, (byte) 0xE8, (byte) 0xA5, (byte) 0x38, (byte) 0x78, (byte) 0x5B,
        (byte) 0x3C, (byte) 0x56, (byte) 0x9A, (byte) 0x2E, (byte) 0x1A, (byte) 0x57,
        (byte) 0x0C, (byte) 0xDD, (byte) 0xF8, (byte) 0x40, (byte) 0x9E, (byte) 0x27,
        (byte) 0xD8, (byte) 0xC3, (byte) 0xB1, (byte) 0xD4, (byte) 0x61, (byte) 0xDC,
        (byte) 0xF9, (byte) 0x45, (byte) 0xE2, (byte) 0x4F, (byte) 0x42, (byte) 0xA5,
        (byte) 0x6E, (byte) 0xE1, (byte) 0x37, (byte) 0x53, (byte) 0x5E, (byte) 0x8C,
        (byte) 0x07, (byte) 0x29, (byte) 0x86, (byte) 0x5B, (byte) 0x38, (byte) 0x53,
        (byte) 0xE8, (byte) 0xA4, (byte) 0x50, (byte) 0x00, (byte) 0x00, (byte) 0x00,
        (byte) 0x00, (byte) 0x49, (byte) 0x45, (byte) 0x4E, (byte) 0x44, (byte) 0xAE,
        (byte) 0x42, (byte) 0x60, (byte) 0x82
      };
    byte imBack[] =
      {
        (byte) 0x89, (byte) 0x50, (byte) 0x4E, (byte) 0x47, (byte) 0x0D, (byte) 0x0A,
        (byte) 0x1A, (byte) 0x0A, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x0D,
        (byte) 0x49, (byte) 0x48, (byte) 0x44, (byte) 0x52, (byte) 0x00, (byte) 0x00,
        (byte) 0x00, (byte) 0x0A, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x0A,
        (byte) 0x08, (byte) 0x06, (byte) 0x00, (byte) 0x00, (byte) 0x01, (byte) 0xFA,
        (byte) 0x35, (byte) 0xFF, (byte) 0x2B, (byte) 0x00, (byte) 0x00, (byte) 0x00,
        (byte) 0x04, (byte) 0x67, (byte) 0x41, (byte) 0x4D, (byte) 0x41, (byte) 0x00,
        (byte) 0x00, (byte) 0xB1, (byte) 0x8F, (byte) 0x0B, (byte) 0xFC, (byte) 0x61,
        (byte) 0x05, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x06, (byte) 0x62,
        (byte) 0x4B, (byte) 0x47, (byte) 0x44, (byte) 0x00, (byte) 0xFF, (byte) 0x00,
        (byte) 0xFF, (byte) 0x00, (byte) 0xFF, (byte) 0xA0, (byte) 0xBD, (byte) 0xA7,
        (byte) 0x93, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x09, (byte) 0x70,
        (byte) 0x48, (byte) 0x59, (byte) 0x73, (byte) 0x00, (byte) 0x00, (byte) 0x0B,
        (byte) 0x13, (byte) 0x00, (byte) 0x00, (byte) 0x0B, (byte) 0x13, (byte) 0x01,
        (byte) 0x00, (byte) 0x9A, (byte) 0x9C, (byte) 0x18, (byte) 0x00, (byte) 0x00,
        (byte) 0x00, (byte) 0x91, (byte) 0x49, (byte) 0x44, (byte) 0x41, (byte) 0x54,
        (byte) 0x18, (byte) 0xD3, (byte) 0x7D, (byte) 0x8F, (byte) 0xD1, (byte) 0x0D,
        (byte) 0xC3, (byte) 0x20, (byte) 0x10, (byte) 0x43, (byte) 0xDF, (byte) 0x45,
        (byte) 0x4C, (byte) 0xC2, (byte) 0x2C, (byte) 0x59, (byte) 0x81, (byte) 0x15,
        (byte) 0xB2, (byte) 0x13, (byte) 0x19, (byte) 0x01, (byte) 0x56, (byte) 0xE8,
        (byte) 0x2A, (byte) 0xB9, (byte) 0x55, (byte) 0xDC, (byte) 0x0F, (byte) 0xB8,
        (byte) 0x26, (byte) 0xB4, (byte) 0x52, (byte) 0x8D, (byte) 0x2C, (byte) 0x99,
        (byte) 0x93, (byte) 0xCF, (byte) 0x18, (byte) 0x93, (byte) 0x04, (byte) 0xC0,
        (byte) 0xD6, (byte) 0x4F, (byte) 0x13, (byte) 0x6E, (byte) 0xA2, (byte) 0x55,
        (byte) 0x84, (byte) 0x2E, (byte) 0x42, (byte) 0xA1, (byte) 0xCF, (byte) 0x45,
        (byte) 0x12, (byte) 0x86, (byte) 0x10, (byte) 0x40, (byte) 0x92, (byte) 0x33,
        (byte) 0xD0, (byte) 0xEA, (byte) 0xF4, (byte) 0x4C, (byte) 0x6F, (byte) 0xAB,
        (byte) 0x68, (byte) 0xE3, (byte) 0x0B, (byte) 0x65, (byte) 0x07, (byte) 0x6B,
        (byte) 0x75, (byte) 0xEC, (byte) 0x02, (byte) 0x94, (byte) 0x43, (byte) 0x06,
        (byte) 0x60, (byte) 0xBA, (byte) 0xEE, (byte) 0x61, (byte) 0x7F, (byte) 0x4D,
        (byte) 0x11, (byte) 0x79, (byte) 0x71, (byte) 0x96, (byte) 0x4C, (byte) 0xF9,
        (byte) 0x20, (byte) 0xC0, (byte) 0x16, (byte) 0x83, (byte) 0x05, (byte) 0x51,
        (byte) 0x69, (byte) 0x65, (byte) 0x56, (byte) 0xFC, (byte) 0x22, (byte) 0x98,
        (byte) 0x86, (byte) 0x3D, (byte) 0x03, (byte) 0x60, (byte) 0xD9, (byte) 0x67,
        (byte) 0xBC, (byte) 0x53, (byte) 0xF6, (byte) 0x0C, (byte) 0x6E, (byte) 0x8A,
        (byte) 0x52, (byte) 0xD3, (byte) 0xE8, (byte) 0x58, (byte) 0xBE, (byte) 0x5F,
        (byte) 0x19, (byte) 0xDA, (byte) 0x79, (byte) 0x56, (byte) 0x48, (byte) 0xCF,
        (byte) 0x92, (byte) 0x61, (byte) 0xFA, (byte) 0xE9, (byte) 0x07, (byte) 0x98,
        (byte) 0x24, (byte) 0xFA, (byte) 0x69, (byte) 0xE2, (byte) 0x0F, (byte) 0xCA,
        (byte) 0x21, (byte) 0x7B, (byte) 0x03, (byte) 0x8F, (byte) 0x5D, (byte) 0x74,
        (byte) 0xC9, (byte) 0xC2, (byte) 0x63, (byte) 0x68, (byte) 0xA1, (byte) 0x00,
        (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x49, (byte) 0x45, (byte) 0x4E,
        (byte) 0x44, (byte) 0xAE, (byte) 0x42, (byte) 0x60, (byte) 0x82
      };
    private String currDirName;
    private static final String dbName = "___org.pabloid.ui.FileBrowser___";
    private String currFile;
    private CommandListener cmdListener;
    private Command cmdNewFile = new Command("Новый файл", Command.OK, 0);
    private Command cmdNewFolder = new Command("Новая папка", Command.OK, 0);
    private Command cmdUp = new Command("Вверх (..)", Command.BACK, 0);
    private Command cmdDelete = new Command("Удалить", Command.OK, 0);
    private Command cmdOk = new Command("OK", Command.OK, 0);
    private Command cmdBack = new Command("Назад", Command.BACK, 0);
    private final Image dirIcon, backIcon;
    private TextBox tbFile;
    private TextBox tbFolder;
    private PList lDelete = new PList();
    private Hashtable htIcons = new Hashtable();

    /* special string denotes upper directory */
    private static final String UP_DIRECTORY = "..";

    /* special string that denotes upper directory accessible by this browser.
     * this virtual directory contains all roots.
     */
    private static final String MEGA_ROOT = "/";

    /* separator string as defined by FC specification */
    private static final String SEP_STR = "/";

    /* separator character as defined by FC specification */
    private static final char SEP = '/';
    private Display display;
    private String selectedURL;
    private String filter = null;
    private String title;
    private File currDir = null;

    /**
     * Creates a new instance of FileBrowser for given <code>Display</code> object.
     * @param display non null display object.
     */
    public FileBrowser(Display display)
      {
        super();
        this.display = display;
        dirIcon = Image.createImage(imDir, 0, imDir.length);
        backIcon = Image.createImage(imBack, 0, imBack.length);
        try
          {
            currDirName = loadDir();
            super.setCommandListener(this);
            addCommand(cmdNewFile);
            addCommand(cmdNewFolder);
            addCommand(cmdUp);
            addCommand(cmdDelete);
            addCommand(null);
            tbFile = new TextBox("Имя файла", "", 128, TextField.ANY);
            tbFile.addCommand(cmdOk);
            tbFile.addCommand(cmdBack);
            tbFile.setCommandListener(this);
            tbFolder = new TextBox("Имя папки", "", 128, TextField.ANY);
            tbFolder.addCommand(cmdOk);
            tbFolder.addCommand(cmdBack);
            tbFolder.setCommandListener(this);
            lDelete.append("Нет");
            lDelete.append("Да");
            lDelete.setCommandListener(this);
            showDir();
          }
        catch (Throwable t)
          {
            alert("FileBrowser()", t);
          }

      }

    public String getCurrDir()
      {
        return currDirName;
      }

    private String loadDir()
      {
        try
          {
            RecordStoreInputStream ris = new RecordStoreInputStream(dbName);
            DataInputStream dis = new DataInputStream(ris);
            String s = dis.readUTF();
            dis.close();
            return s;
          }
        catch (Throwable e)
          {
            return MEGA_ROOT;
          }
      }

    private void saveDir(String dir)
      {
        try
          {
            RecordStoreOutputStream ros = new RecordStoreOutputStream(dbName);
            DataOutputStream dos = new DataOutputStream(ros);
            dos.writeUTF(dir);
            dos.close();
          }
        catch (Throwable e)
          {
            alert("saveDir()", e);
          }
      }

    public int compare(Object o1, Object o2)
      {
        String s1 = o1.toString();
        String s2 = o2.toString();
        if (s1.endsWith(SEP_STR) && !s2.endsWith(SEP_STR))
            return -1;
        else if (s2.endsWith(SEP_STR) && !s1.endsWith(SEP_STR))
            return 1;
        else
            return s1.toLowerCase().compareTo(s2.toLowerCase());
      }

    public final void showDir()
      {
        new Thread(new Runnable()
          {

            public void run()
              {
                try
                  {
                    showCurrDir();
                  }
                catch (Throwable e)
                  {
                    e.printStackTrace();
                    alert("showDir()", e);
                  }
              }
          }).start();
      }

    private void delete(final String url)
      {
        new Thread(new Runnable()
          {

            public void run()
              {
                try
                  {
                    File f = File.open(url);
                    f.delete();
                    f.close();
                  }
                catch (Throwable t)
                  {
                    t.printStackTrace();
                    alert("delete()", t);
                  }
              }
          }).start();
      }

    private void alert(String s, Throwable t)
      {
        t.printStackTrace();
        Alert alert = new Alert("Ошибка", s + " " + t, null, AlertType.ERROR);
        alert.setTimeout(-2);
        display.setCurrent(alert, FileBrowser.this);
      }

    /**
     * Indicates that a command event has occurred on Displayable d.
     * @param command a <code>Command</code> object identifying the command. This is either
     * one of the applications have been added to <code>Displayable</code> with <code>addCommand(Command)</code>
     * or is the implicit <code>SELECT_COMMAND</code> of PList.
     * @param d the <code>Displayable</code> on which this event has occurred
     */
    public void commandAction(Command command, Displayable d)
      {
        if (d == tbFile)
          {
            if (command == cmdOk)
                new Thread(new Runnable()
                  {

                    public void run()
                      {
                        try
                          {
                            File f = File.open(currDirName + tbFile.getString());
                            if (!f.exists())
                                f.create();
                            f.close();
                            showDir();
                          }
                        catch (Throwable t)
                          {
                            t.printStackTrace();
                            alert("createFile()", t);
                          }
                      }
                  }).start();
            display.setCurrent(this);
          }
        else if (d == tbFolder)
          {
            if (command == cmdOk)
                new Thread(new Runnable()
                  {

                    public void run()
                      {
                        try
                          {
                            File f = File.open(currDirName + tbFolder.getString() + (tbFolder.getString().endsWith("/") ? "" : "/"));
                            if (!f.exists())
                                f.mkdir();
                            f.close();
                            showDir();
                          }
                        catch (Throwable t)
                          {
                            t.printStackTrace();
                            alert("createFolder()", t);
                          }
                      }
                  }).start();
            display.setCurrent(this);
          }
        else if (d == lDelete)
          {
            if (lDelete.getSelectedIndex() == 1)
                delete(currDirName + getString(getSelectedIndex()));
            showDir();
            display.setCurrent(this);
          }
        else if (command.equals(SELECT_COMMAND))
          {
            PList curr = (PList) d;
            currFile = curr.getString(curr.getSelectedIndex());
            new Thread(new Runnable()
              {

                public void run()
                  {
                    if (currFile.endsWith(SEP_STR) || currFile.equals(UP_DIRECTORY))
                        openDir(currFile);
                    else
                        //switch To Next
                        doDismiss();
                  }
              }).start();
          }
        else if (command.equals(cmdNewFile))
            display.setCurrent(tbFile);
        else if (command.equals(cmdNewFolder))
            display.setCurrent(tbFolder);
        else if (command.equals(cmdDelete))
          {
            lDelete.setTitle("Удалить " + getString(getSelectedIndex()) + "?");
            lDelete.setSelectedIndex(0, true);
            display.setCurrent(lDelete);
          }
        else if (command.equals(cmdUp))
            new Thread(new Runnable()
              {

                public void run()
                  {
                    openDir(UP_DIRECTORY);
                  }
              }).start();
        else
            cmdListener.commandAction(command, d);
      }

    /**
     * Sets component's title.
     *  @param title component's title.
     */
    public void setTitle(String title)
      {
        this.title = title;
        super.setTitle(title);
      }

    /**
     * Show file list in the current directory .
     */
    private void showCurrDir()
      {
        if (title == null)
            super.setTitle(currDirName);
        int dbg = 0;
        String[] arr = null;
        dbg++;//0
        try
          {
            synchronized (this)
              {
                deleteAll();
                dbg++;//1

                currDir = File.open(currDirName);
                dbg++;//2
                if (!currDir.exists() || currDirName.equals(MEGA_ROOT))
                  {
                    currDirName = MEGA_ROOT;
                    dbg++;//3
                    currDir = File.open(currDirName);
                    dbg++;//4
                  }
                arr = currDir.list();
                dbg++;//5

                if (!currDirName.equals(MEGA_ROOT))
                    append(UP_DIRECTORY, getIcon(UP_DIRECTORY));
                if (arr != null)
                  {
                    dbg++;//6
                    Arrays.sort(arr, this);
                    dbg++;//7
                    for (int i = 0; i < arr.length; i++)
                      {
                        String fileName = arr[i];
                        dbg++;//8
                        if (fileName.charAt(fileName.length() - 1) == SEP)
                            // This is directory
                            append(fileName, getIcon(fileName));
                        else // this is regular file
                          {
                            RE fil = new RE(".*");
                            dbg++;//9
                            if (filter != null)
                                fil = new RE(RE.simplePatternToFullRegularExpression(filter.replace(';', '|')));
                            if (fil.match(fileName))
                                append(fileName, getIcon(fileName));
                            dbg++;//10
                          }
                      }
                  }

                if (currDir != null)
                    currDir.close();
                dbg++;//11

              }
          }
        catch (Throwable t)
          {
            alert("showCurrDir(" + dbg + ") " + Arrays.toString(arr) + " " + currDir, t);
          }
      }

    public void addExt(String ext, Image icon)
      {
        htIcons.put(ext, icon);
      }

    public Image getIcon(String filename)
      {
        if (filename.endsWith(SEP_STR))
            return dirIcon;
        if (filename.equals(UP_DIRECTORY))
            return backIcon;
        String ext = filename.substring(Math.max(filename.lastIndexOf('.') + 1, filename.lastIndexOf(SEP) + 1));
        Image icon = (Image) htIcons.get(ext);
        if (icon != null)
            return icon;
        Font fnt = Font.getDefaultFont();
        int ch = fnt.getHeight() + 2;
        ext = ext.substring(0, Math.min(ext.length(), 3)).toLowerCase();
        int cw = fnt.stringWidth(ext) + 2;
        icon = Image.createImage(cw, ch);
        Graphics g = icon.getGraphics();
        int[] cl =
          {
            0, 0, 0
          };
        for (int i = 0; i < 3 && i < ext.length(); i++)
            cl[i] = getLetterColor(ext.charAt(i));
        int avg = (cl[0] + cl[1] + cl[2]) / 3;
        if (avg >= 128)
            g.setColor(255 - cl[0], 255 - cl[1], 255 - cl[2]);
        else
            g.setColor(cl[0], cl[1], cl[2]);
        g.setFont(fnt);
        g.drawString(ext, 1, 1, 20);
        return icon;
      }

    private int getLetterColor(int letter)
      {
        return Math.abs(((255 * (letter - 'a')) / ('z' - 'a'))) % 255;
      }

    private void openDir(String fileName)
      {
        /* In case of directory just change the current directory
         * and show it
         */
        if (currDirName.equals(MEGA_ROOT))
          {
            if (fileName.equals(UP_DIRECTORY))
                // can not go up from MEGA_ROOT
                return;
            currDirName = fileName;


          }
        else if (fileName.equals(UP_DIRECTORY))
          {
            // Go up one directory
            // TODO use setFileConnection when implemented
            int i = currDirName.lastIndexOf(SEP, currDirName.length() - 2);


            if (i != -1)
                currDirName = currDirName.substring(0, i + 1);
            else
                currDirName = MEGA_ROOT;
          }
        else
            currDirName += fileName;
        showDir();


      }

    /**
     * Returns selected file as a <code>FileConnection</code> object.
     * @return non null <code>FileConection</code> object
     */
    public File getSelectedFile() throws IOException
      {
        File fileConnection = File.open(selectedURL);
        return fileConnection;


      }

    /**
     * Returns selected <code>FileURL</code> object.
     * @return non null <code>FileURL</code> object
     */
    public String getSelectedFileURL()
      {
        return selectedURL;


      }

    /**
     * Sets the file filter.
     * @param filter file filter String object
     */
    public void setFilter(String filter)
      {
        this.filter = filter;


      }

    /**
     * Returns command listener.
     * @return non null <code>CommandListener</code> object
     */
    protected CommandListener getCommandListener()
      {
        return cmdListener;


      }

    /**
     * Sets command listener to this component.
     * @param commandListener <code>CommandListener</code> to be used
     */
    public void setCommandListener(CommandListener commandListener)
      {
        this.cmdListener = commandListener;


      }

    private void doDismiss()
      {
        selectedURL = "file:///" + currDirName + currFile;
        CommandListener cmdLis = getCommandListener();


        if (cmdLis != null)
            cmdLis.commandAction(SELECT_COMMAND, this);
        saveDir(
                currDirName);

      }
  }
